"use strict";

var _graphql = require("../../graphql");

var _ = require("../..");

var _ObjectTypeComposer = require("../../ObjectTypeComposer");

var _InputTypeComposer = require("../../InputTypeComposer");

var _InterfaceTypeComposer = require("../../InterfaceTypeComposer");

var _toInputObjectType = require("../toInputObjectType");

describe('toInputObjectType()', () => {
  let PersonTC;
  beforeEach(() => {
    _.schemaComposer.clear();

    PersonTC = _ObjectTypeComposer.ObjectTypeComposer.create({
      name: 'Person',
      fields: {
        name: 'String',
        age: {
          type: _graphql.GraphQLInt
        },
        address: {
          type: new _graphql.GraphQLObjectType({
            name: 'Address',
            fields: {
              city: {
                type: _graphql.GraphQLString
              },
              street: {
                type: _graphql.GraphQLString
              }
            }
          })
        }
      }
    }, _.schemaComposer);
  });
  it('should return InputTypeComposer', () => {
    const itc = (0, _toInputObjectType.toInputObjectType)(PersonTC);
    expect(itc).toBeInstanceOf(_InputTypeComposer.InputTypeComposer);
    expect(itc.getTypeName()).toBe('PersonInput');
  });
  it('should accept prefix in opts', () => {
    const itc = (0, _toInputObjectType.toInputObjectType)(PersonTC, {
      prefix: 'SomePrefix'
    });
    expect(itc.getTypeName()).toBe('SomePrefixPersonInput');
  });
  it('should accept postfix in opts', () => {
    const itc = (0, _toInputObjectType.toInputObjectType)(PersonTC, {
      postfix: 'PostfixInpt'
    });
    expect(itc.getTypeName()).toBe('PersonPostfixInpt');
  });
  it('should keep scalar types', () => {
    const itc = (0, _toInputObjectType.toInputObjectType)(PersonTC);
    expect(itc.getFieldType('name')).toBe(_graphql.GraphQLString);
    expect(itc.getFieldType('age')).toBe(_graphql.GraphQLInt);
  });
  it('should convert field with ObjectType to InputType', () => {
    const itc = (0, _toInputObjectType.toInputObjectType)(PersonTC);
    const addrType = itc.getFieldType('address');
    expect(addrType).toBeInstanceOf(_graphql.GraphQLInputObjectType);
    expect(addrType.name).toBe('AddressInput');
  });
  it('should reuse generated input type for recursive types', () => {
    PersonTC.setField('spouce', PersonTC);
    const itc = (0, _toInputObjectType.toInputObjectType)(PersonTC);
    expect(itc.getFieldType('spouce')).toBe(itc.getType());
  });
  it('should reuse generated input type for recursive types in List', () => {
    PersonTC.setField('friends', PersonTC.getTypePlural());
    const itc = (0, _toInputObjectType.toInputObjectType)(PersonTC);
    expect(itc.getFieldType('friends')).toBeInstanceOf(_graphql.GraphQLList);
    expect(itc.getFieldType('friends').ofType).toBe(itc.getType());
  });
  it('should convert InterfaceTypeComposer to InputTypeComposer', () => {
    const iftc = _InterfaceTypeComposer.InterfaceTypeComposer.create(`
      interface IFace {
        name: String
        age: Int
      }
    `, _.schemaComposer);

    const itc = (0, _toInputObjectType.toInputObjectType)(iftc);
    expect(itc.getFieldType('name')).toBe(_graphql.GraphQLString);
    expect(itc.getFieldType('age')).toBe(_graphql.GraphQLInt);
    expect(itc.getTypeName()).toBe('IFaceInput');
  });
  it('should convert field with InterfaceType to InputType', () => {
    _InterfaceTypeComposer.InterfaceTypeComposer.create(`
      interface IFace {
        name: String
        age: Int
      }
    `, _.schemaComposer);

    const tc = _ObjectTypeComposer.ObjectTypeComposer.create(`
      type Example implements IFace {
        name: String
        age: Int
        neighbor: IFace
      }
    `, _.schemaComposer);

    const itc = (0, _toInputObjectType.toInputObjectType)(tc);
    expect(itc.getFieldType('name')).toBe(_graphql.GraphQLString);
    expect(itc.getFieldType('age')).toBe(_graphql.GraphQLInt);
    const ifaceField = itc.getFieldTC('neighbor');
    expect(ifaceField).toBeInstanceOf(_InputTypeComposer.InputTypeComposer);

    if (ifaceField instanceof _InputTypeComposer.InputTypeComposer) {
      expect(ifaceField.getType()).toBeInstanceOf(_graphql.GraphQLInputObjectType);
      expect(ifaceField.getTypeName()).toBe('IFaceInput');
      expect(ifaceField.getFieldType('name')).toBe(_graphql.GraphQLString);
      expect(ifaceField.getFieldType('age')).toBe(_graphql.GraphQLInt);
      expect(itc.getTypeName()).toBe('ExampleInput');
    }
  });
});